import json
from ansible.errors import AnsibleError as ae
from ansible.plugins.action import ActionBase

import requests

try:
    from __main__ import display
except ImportError:
    from ansible.utils.display import Display
    display = Display()

class ActionModule(ActionBase):
    ''' Manage hosts and infrastructure at JiffyBox'''

    ### We need to be able to modify the inventory
    BYPASS_HOST_LOOP = True
    TRANSFERS_FILES = False

    def __init__(self, task, connection, play_context, loader, templar,
                 shared_loader_obj):
        super(ActionModule, self).__init__(task, connection, play_context,
                                           loader, templar, shared_loader_obj)
        self.api_token = None
        self.url = None
        self.args = {}
        self.vars = {}

        self.devices = []
        self.backups = []
        self.tarifs = []
        self.distros= []
        self.ips = []
        self.checks = []
        self.contacts = []
        self.documents = []

    def run(self, tmp=None, task_vars=None):
        ''' handler for file transfer operations '''
        if task_vars is None:
            task_vars = dict()
        self.vars = task_vars

        result = super(ActionModule, self).run(tmp, task_vars)

        if self._play_context.check_mode:
            result['skipped'] = True
            result['msg'] = 'check mode not supported for this module'
            return result

        self.api_token = self._task.args.get('api_token', task_vars.get('jiffy_api_token', False))
        if not self.api_token:
            raise ae("'api_token' is a required argument or you define the variable 'jiffy_api_token' in your inventory.")

        action = self._task.args.get('action', 'list')

        if action == 'list':
            self.list_devices()
        elif action == 'moveip':
            self.move_ip()
        else:
            raise ae("'%s' is an unknown action" % action)

        result['changed'] = False
        return result

    def _request(self, path, data = None, method = 'GET'):
        encoder = json.JSONEncoder()
        postData = {}

        if self.url is None:
            self.url = 'https://api.jiffybox.de/' + self.api_token + '/v1.0/'

        if data:
            method = 'POST'
            for key in data:
                item = data.get(key)
                if type(item) is list or type(item) is dict:
                    if len(item) > 0:
                        item = encoder.encode(item)
                if type(item) is int or type(item) is unicode or type(item) is bool:
                    item = str(item)
                if item and type(item) is str and len(item) > 0:
                    postData.__setitem__(key, item)

        request_result = {}
        try:
            if method == 'GET':
                request_result = requests.get(self.url + path)
            elif method == 'POST':
                request_result = requests.put(self.url + path, data = postData)
            elif method == 'DELETE':
                request_result = requests.delete(self.url + path)
        except ae as e:
            raise ae('No result from JiffyBox API')

        decoder = json.JSONDecoder()
        content = decoder.decode(request_result.content.decode('utf8'))
        if not content['result']:
            msg = content['messages']
            raise ae('%s' % msg)
        return content['result']

    def _load_objects(self, type, path):
        display.vv("Reading %s from JiffyBox" % type)
        changed = False
        if '_jiffybox_' + type not in self.vars:
            changed = True
            self.vars.__setitem__('_jiffybox_' + type, self._request(path))
        return changed, self.vars.get('_jiffybox_' + type)

    def load_devices(self):
        if len(self.devices) == 0:
            (changed, self.devices) = self._load_objects('devices', 'jiffyBoxes')

    def load_ips(self):
        if len(self.ips) == 0:
            (changed, self.ips) = self._load_objects('ips', 'ips')

    def find_host(self, name):
        self.load_devices()
        for id in self.devices:
            if self.devices[id].get('name') == name:
                return self.devices[id]
        return False

    def find_host_by_ip(self, ip):
        self.load_devices()
        self.load_ips()
        for block in self.ips:
            if str(ip.get('id')) in self.ips[block]:
                return self.devices[block]
        return False

    def find_host_in_group(self, group, excluded):
        for name in group:
            if name != excluded:
                return self.find_host(name)
        return False

    def find_ip(self, address):
        self.load_ips()
        for block in self.ips:
            for ip in self.ips[block]:
                if self.ips[block][ip].get('ip') == address:
                    return self.ips[block][ip]
        return False

    def find_floating_ip(self):
        self.load_ips()
        selected = self._task.args.get('ip', False)
        ips = []
        for block in self.ips:
            for ip in self.ips[block]:
                if self.ips[block][ip].get('floating') == 'true':
                    if ip == selected:
                        return self.ips[block][ip]
                    ips.append(self.ips[block][ip])
        if len(ips) == 0:
            raise ae('There is no floating ip address available.')

        return ips

    def list_devices(self):
        self.load_devices()
        columns = {'id': 10, 'name': 10, 'ips': 20}
        rows = []
        exist_floating = False
        for device in self.devices:
            id = str(self.devices[device].get('id'))
            name = self.devices[device].get('name')
            f_ips = []
            if name in self.vars.get('ansible_play_hosts'):
                for address in self.devices[device].get('ips').get('public'):
                    ip = self.find_ip(address)
                    if ip:
                        address += ' [' + str(ip.get('id')) + ']'
                        if ip.get('floating') == 'true':
                            address += '*'
                            exist_floating = True
                    f_ips.append(address)
                ips = ' '.join(f_ips)
                if len(id) > columns['id']:
                    columns['id'] = len(id)
                if len(name) > columns['name']:
                    columns['name'] = len(name)
                if len(ips) > columns['ips']:
                    columns['ips'] = len(ips)
                rows.append([name, id, ips])
        rows.sort()
        rows.insert(0, ['Hostname', 'ID', 'IP [ID]'])
        rows.insert(1, [''.ljust(columns['name'], '-'), ''.ljust(columns['id'], '-'), ''.ljust(columns['ips'], '-')])
        output = "\n"
        for row in rows:
            output += row[0].ljust(columns['name']) + ' ' + row[1].ljust(columns['id']) + ' ' + row[2].ljust(columns['ips']) + "\n"
        if exist_floating:
            output += "\n* This marks a floating ip address\n"
        print(output)
        return ''

    def move_ip(self):
        display.vv("Moving IP ...")
        target = self._task.args.get('target', False)
        group = self._task.args.get('group', False)
        if not target and not group:
            raise ae("'target' is a required argument.")
        if target:
            targethost = self.find_host(target)
            if not targethost:
                raise ae("Target host %s unknown in this JiffyBox account." % target)
        else:
            targethost = False

        ips = self.find_floating_ip()
        for ip in ips:
            if not ip:
                raise ae('No ip defined for moving.')
            display.vv("- floating IP: %s" % ip.get('ip'))

            sourcehost = self.find_host_by_ip(ip)
            if not sourcehost:
                display.vv('Current host for the floating ip can not be found.')
            else:
                display.vv("- source host: %s" % sourcehost.get('name'))

                if not target:
                    targethost = self.find_host_in_group(group, sourcehost.get('name'))

                if not targethost:
                    pass
                elif sourcehost.get('id') == targethost.get('id'):
                    display.vv('- cancelled as source and target are the same')
                else:
                    display.vv("- target host: %s" % targethost.get('name'))
                    path = 'ips/' + str(sourcehost.get('id')) + '/' + str(ip.get('id')) + '/move'
                    self._request(path, {'targetid': targethost.get('id')})
